/*
** Copyright (C) 1998,1999 Martin Roesch <roesch@clark.net>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "log.h"

extern PrintIP pip;  /* we need this to log properly */
char *data_dump_buffer;  /* printout buffer for PrintNetData */
int dump_ready;      /* flag to indicate status of printout buffer */
int dump_size;


/****************************************************************************
 *
 * Function: OpenLogFile()
 *
 * Purpose: Create the log directory and file to put the packet log into
 *
 * Arguments: None.
 *
 * Returns: 0 on success, exits on error
 *
 ****************************************************************************/
int OpenLogFile()
{
   char log_path[STD_BUF]; /* path to log file */
   char log_file[STD_BUF]; /* name of log file */
   char proto[5];          /* logged packet protocol */


   /* zero out our buffers */
   bzero(log_path, STD_BUF);
   bzero(log_file, STD_BUF);
   bzero(proto, 5);

   /* figure out which way this packet is headed in relation to the homenet */
   if((net.dip & pv.netmask) == pv.homenet)
   {
      if((net.sip & pv.netmask) != pv.homenet)
      {
         sprintf(log_path, "%s/%s", pv.log_dir, pip.saddr);
      }
      else
      {
         if( net.sp >= net.dp )
         {
            sprintf(log_path, "%s/%s", pv.log_dir, pip.saddr);
         }
         else
         {
            sprintf(log_path, "%s/%s", pv.log_dir, pip.daddr);
         }
      }
   }
   else
   {
      if((net.sip & pv.netmask) == pv.homenet)
      {
         sprintf(log_path, "%s/%s", pv.log_dir, pip.daddr);
      }
      else
      {
         if(net.sp >= net.dp)
         {
            sprintf(log_path, "%s/%s", pv.log_dir, pip.saddr);
         }
         else
         {
            sprintf(log_path, "%s/%s", pv.log_dir, pip.daddr);
         }
      }
   }

#ifdef DEBUG
   fprintf(stderr, "Creating directory: %s\n",log_path);
#endif

   /* build the log directory */
   if(mkdir(log_path,S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH))
   {
#ifdef DEBUG
      if(errno != EEXIST)
      {
         printf("Problem creating directory %s\n",log_path);
      }
#endif
   }

#ifdef DEBUG
   printf("Directory Created!\n");
#endif

   /* build the log filename */
   if((!strcasecmp(pip.proto, "TCP"))||
      (!strcasecmp(pip.proto, "UDP")))
   {
      if(pip.sport >= pip.dport)
      {
         sprintf(log_file, "%s/%s:%d-%d", log_path, pip.proto, pip.sport,
                 pip.dport);
      }
      else
      {
         sprintf(log_file, "%s/%s:%d-%d", log_path, pip.proto, pip.dport,
                 pip.sport);
      }
   }
   else
   {
      if(!strcasecmp(pip.proto, "ICMP"))
      {
         sprintf(log_file, "%s/%s_%s", log_path, pip.proto, pip.icmp_str);
      }
      else
      {
         sprintf(log_file, "%s/IP_FRAG", log_path);
      }
   }

#ifdef DEBUG
   printf("Opening file: %s\n", log_file);
#endif

   /* finally open the log file */
   if((log_ptr = fopen(log_file, "a")) == NULL)
   {
       fprintf(stderr, "ERROR: OpenLogFile() => fopen() log file: %s\n",
               strerror(errno));
       exit(1);
   }

#ifdef DEBUG
   printf("File opened...\n");
#endif

   return 0;
}



/****************************************************************************
 *
 * Function: PrintNetData(FILE *, char *,int)
 *
 * Purpose: Do a side by side dump of a buffer, hex dump of buffer bytes on
 *          the left, decoded ASCII on the right.
 *
 * Arguments: fp => ptr to stream to print to
 *            start => pointer to buffer data
 *            len => length of data buffer
 *
 * Returns: void function
 *
 ****************************************************************************/
void PrintNetData(FILE *fp, char *start, const int len)
{
   char *end;              /* ptr to buffer end */
   int i;                  /* counter */
   int j;                  /* counter */
   int dbuf_size;          /* data buffer size */
   int done;               /* flag */
   char *data;             /* index pointer */
   char *frame_ptr;        /* we use 66 byte frames for a printed line */
   char *d_ptr;            /* data pointer into the frame */
   char *c_ptr;            /* char pointer into the frame */
   char conv[] = "0123456789ABCDEF"; /* xlation lookup table */

   /* initialization */
   done = 0;

   /* zero, print a <CR> and get out */
   if(!len)
   {
      fputc('\n', fp);
      /*fflush(fp);*/
      return;
   }

   if(dump_ready)
   {
      fwrite(data_dump_buffer, dump_size, 1, fp);
      fputc('\n', fp);
      /*fflush(fp);*/
      return;
   }

   end = start + (len-1); /* set the end of buffer ptr */

   dbuf_size = ((len/16) * 66) + 67;

   data_dump_buffer = (char *) malloc(dbuf_size);
/*   bzero(data_dump_buffer, dbuf_size); */
   memset(data_dump_buffer, 0x20, dbuf_size);
   if(data_dump_buffer == NULL)
   {
      perror("PrintNetData()");
      CleanExit();
   }

   /* set the byte buffer pointer to step thru the data buffer */
   data = start;

   /* set the frame pointer to the start of the printout buffer */
   frame_ptr = data_dump_buffer;

   /* loop thru the whole buffer */
   while(!done)
   {
      /* initialize counters and frame index pointers */
      i = 0;
      j = 0;
      d_ptr = frame_ptr;
      c_ptr = (frame_ptr + C_OFFSET);

      /* process 16 bytes per frame */
      for(i=0; i<16; i++)
      {
         /* look up the ASCII value of the first nybble of the 
            current data buffer */
         *d_ptr = conv[((*data&0xFF) >> 4)];
         d_ptr++;

         /* look up the second nybble */
         *d_ptr = conv[((*data&0xFF)&0x0F)];
         d_ptr++;

         /* put a space in between */
         *d_ptr = 0x20;
         d_ptr++;

         /* print out the char equivalent */
         if(*data > 0x1F && *data < 0x7E)
            *c_ptr = (*data&0xFF);
         else
            *c_ptr = 0x2E;

         c_ptr++;
 
         /* increment the pointer or finish up */
         if(data < end)
            data++;
         else
         {
            done = 1;
            dump_ready = 1;

            *c_ptr='\n';
            c_ptr++;
            *c_ptr = 0;

            dump_size = (int) (c_ptr - data_dump_buffer);
            fwrite(data_dump_buffer, dump_size, 1, fp);
            fputc('\n', fp);
            fflush(fp);
            return;
         }
      }
      *c_ptr = '\n';
      frame_ptr += FRAME_SIZE;
   }
}



/****************************************************************************
 *
 * Function: PrintIPPkt(FILE *, int)
 *
 * Purpose: Display the packet date to the stream pointer
 *
 * Arguments: fp => pointer to print data to
 *            type => packet protocol
 *
 * Returns: void function
 *
 ****************************************************************************/
void PrintIPPkt(FILE *fp, int type)
{
#ifdef DEBUG
   printf("PrintIPPkt type = %d\n", type);
#endif

   switch(type)
   {
      case IPPROTO_TCP:
                fprintf(fp, "%s %s:%d -> %s:%d %s TTL:%d", pip.timestamp,
                        pip.saddr, pip.sport, pip.daddr, pip.dport, pip.proto, 
                        pip.ttl);

                if(pip.ip_df)
                   fprintf(fp, " DF");

                if(pip.ip_mf)
                   fprintf(fp, " MF");

                if(pip.frag_off & 0x1FFF)
                   fprintf(fp, "\nFragment Offset: %d", (pip.frag_off&0x1FFF));

                fprintf(fp, "\n");

                if(pip.flags & TH_SYN) fprintf(fp, "S"); else fprintf(fp, "*");
                if(pip.flags & TH_FIN) fprintf(fp, "F"); else fprintf(fp, "*");
                if(pip.flags & TH_RST) fprintf(fp, "R"); else fprintf(fp, "*");
                if(pip.flags & TH_PUSH) fprintf(fp, "P"); else fprintf(fp, "*");
                if(pip.flags & TH_ACK) fprintf(fp, "A"); else fprintf(fp, "*");
                if(pip.flags & TH_URG) fprintf(fp, "U"); else fprintf(fp, "*");

                fprintf(fp, " Seq:%lX Ack:%lX Win:%lX\n",pip.seq, pip.ack,
                        pip.win);
                break;

      case IPPROTO_UDP:
                fprintf(fp, "%s %s:%d -> %s:%d %s TTL:%d",
                        pip.timestamp, pip.saddr, pip.sport, pip.daddr, 
                        pip.dport, pip.proto, pip.ttl);

                if(pip.ip_df)
                   fprintf(fp, " DF");

                if(pip.ip_mf)
                   fprintf(fp, " MF");

                if(pip.frag_off & 0x1FFF)
                   fprintf(fp, "\nFragment Offset: %d", (pip.frag_off & 0x1FFF) );

                fprintf(fp, "\n");

                fprintf(fp, "Len: %d\n", pip.udp_len);

                break;

      case IPPROTO_ICMP:
                fprintf(fp, "%s %s -> %s %s TTL:%d", pip.timestamp, pip.saddr, 
                        pip.daddr, pip.proto, pip.ttl); 

                if(pip.ip_df)
                   fprintf(fp, " DF");

                if(pip.ip_mf)
                   fprintf(fp, " MF");
 
                if(pip.frag_off & 0x1FFF )
                   fprintf(fp, "\nFragment Offset: %d", (pip.frag_off & 0x1FFF));

                fprintf(fp, "\n");

                fprintf(fp, "ID:%d Seq:%d %s\n", pip.icmp_id, pip.icmp_seqno, 
                        pip.icmp_str);

                break;
   }
}




void OpenAlertFile()
{
   char filename[STD_BUF];

   sprintf(filename, "%s/alert", pv.log_dir);

   if((alert = fopen(filename, "a")) == NULL)
   {
       fprintf(stderr, "ERROR in OpenAlertFile() => fopen() alert file: %s\n",
               strerror(errno));
       exit(1);
   }
} 



void PrintFragHeader(FILE *fp)
{
   fprintf(fp, "%s %s -> %s  %s TTL:%d", pip.timestamp, pip.saddr,
           pip.daddr, pip.proto, pip.ttl);

   if(pip.ip_df)
      fprintf(fp, " DF");
            
   if(pip.ip_mf)
      fprintf(fp, " MF");
      
   if(pip.frag_off & 0x1FFF )
      fprintf(fp, "\nFragOff: %d", (pip.frag_off & 0x1FFF));
   
   fprintf(fp, "\n");
} 



void ClearDumpBuf()
{
   if(data_dump_buffer != NULL)
      free(data_dump_buffer);

   data_dump_buffer = NULL;
      

   dump_ready = 0;
}
